#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add instance_name field to nodes

Revision ID: 15e9d00367b0
Revises: 1c14278d6e33
Create Date: 2025-06-17 12:59:58.807596

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '15e9d00367b0'
down_revision = '763b2f62d215'


def upgrade():
    op.add_column('nodes', sa.Column('instance_name', sa.String(length=255),
                                     nullable=True))
