#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add dynamic portgroup fields

Revision ID: 2a3b4c5d6e7f
Revises: c1fd28861bb9
Create Date: 2026-03-04 00:00:00.000000

"""

from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '2a3b4c5d6e7f'
down_revision = 'c1fd28861bb9'


def upgrade():
    op.add_column('ports',
                  sa.Column('available_for_dynamic_portgroup',
                            sa.Boolean(),
                            server_default=sa.true(),
                            nullable=False))
    op.add_column('portgroups',
                  sa.Column('dynamic_portgroup',
                            sa.Boolean(),
                            nullable=True))
